//
//  UIViewController+Extensions.swift
//  ThreeDS-Sample
//
//  Created by Drew Pitchford on 5/14/20.
//  Copyright © 2020 mSignia. All rights reserved.
//

import Foundation
import UIKit

extension UIViewController {
    
    // MARK: Injection
    public func assertDependenciesHaveBeenInjected(_ dependencies: [Any?], file: String = #file, function: String = #function) {
        
        for dependency in dependencies {
            
            guard let _ = dependency else {
                
                let okAction = UIAlertAction(title: "Ok", style: .default, handler: { (_) in
                    
                    fatalError("Missing injection in \(file)::\(function)")
                })
                
                showAlert(withTitle: "VC Injection Error", message: "Missing injection in \(file)::\(function). Pressing Ok will kill the app. Please contact the app developer with this info", actions: [okAction])
                return
            }
        }
    }
    
    // MARK: - Setup
    public func setUpNavBarForLargeTitle(using largeTitleDisplayMode: UINavigationItem.LargeTitleDisplayMode = .always) {
        
        if #available(iOS 11.0, *) {
            
            navigationItem.largeTitleDisplayMode = largeTitleDisplayMode
            navigationController?.navigationBar.prefersLargeTitles = true
        }
    }
    
    // MARK: - Show Alerts
    public func showNotYetImplementedAlert() {
        
        showOkAlert(withTitle: "Not Yet Implemented", andMessage: "This feature hasn't been implemented yet.")
    }
    
    public func showAlert(for error: Error, actions: [UIAlertAction]? = nil) {
        
        guard actions == nil else {
            
            showAlert(withTitle: "Web Error", message: (error as NSError).debugDescription, actions: actions)
            return
        }
        
        showOkAlert(withTitle: "Web Error", andMessage: (error as NSError).debugDescription)
    }
    
    private func showOkAlert(withTitle title: String? = nil, andMessage message: String, autoDismissalTime: TimeInterval? = nil) {
        
        let alert = createAlert(withTitle: title, message: message)
        addOkAction(to: alert)
        
        present(alert, animated: true) {
            if let dismissalTime = autoDismissalTime {
                DispatchQueue.main.asyncAfter(deadline: .now() + dismissalTime) {
                    self.dismiss(animated: true, completion: nil)
                }
            }
        }
    }
    
    public func showAlert(withTitle title: String? = nil, message: String, actions: [UIAlertAction]? = nil, autoDismissalTime: TimeInterval? = nil, showOkButton: Bool = true) {
        
        if let actions = actions {
            
            let alert = createAlert(withTitle: title, message: message)
            for action in actions {
                
                alert.addAction(action)
            }
            
            present(alert, animated: true) {
                if let dismissalTime = autoDismissalTime {
                    DispatchQueue.main.asyncAfter(deadline: .now() + dismissalTime) {
                        self.dismiss(animated: true, completion: nil)
                    }
                }
            }
        }
        else if !showOkButton {
            
            let alert = createAlert(withTitle: "", message: message)
            present(alert, animated: true) {
                if let dismissalTime = autoDismissalTime {
                    DispatchQueue.main.asyncAfter(deadline: .now() + dismissalTime) {
                        self.dismiss(animated: true, completion: nil)
                    }
                }
            }
        }
        else {
            
            showOkAlert(withTitle: title, andMessage: message, autoDismissalTime: autoDismissalTime)
        }
    }
    
    // MARK: - Private Alert Helpers
    private func createAlert(withTitle title: String? = nil, message: String) -> UIAlertController {
        
        let alert = UIAlertController(title: title, message: message, preferredStyle: .alert)
        return alert
    }
    
    private func addOkAction(to alert: UIAlertController) {
        
        let okAction = UIAlertAction(title: "Ok", style: .default, handler: nil)
        alert.addAction(okAction)
    }
}

// MARK: - PopoverDelegate
extension UIViewController: UIPopoverPresentationControllerDelegate {
    
    // for iOS 8 - 9.2
    open func adaptivePresentationStyle(for controller: UIPresentationController) -> UIModalPresentationStyle {
        return .none
    }
    
    // for iOS 9.3 and up
    open func adaptivePresentationStyle(for controller: UIPresentationController, traitCollection: UITraitCollection) -> UIModalPresentationStyle {
        return .none
    }
}
