//
//  PureInterfaceSCA.swift
//  ThreeDS-Sample
//
//  Created by Drew Pitchford on 7/1/21.
//  Copyright © 2021 mSignia. All rights reserved.
//

import Foundation
import UIKit
import uSDK
import LocalAuthentication

class PureInterfaceSCATask {
    // MARK: - Properties
    private var accoutNumber: String
    private var viewController: UIViewController
    private var purchaseInfo: PurchaseInformation
    private var transaction: UTransaction! // Best practice is not to do implicity unwrapped optionals; could not think of a better way to handle this though
    private var progressDialog: UProgressDialog?
    private var spinner: UIActivityIndicatorView?
    private var selectedDS: SADirectoryServer?
    
    // MARK: - Init
    init(accountNumber: String, viewController: UIViewController, purchaseInfo: PurchaseInformation, spinner: UIActivityIndicatorView? = nil) {
        self.accoutNumber = accountNumber
        self.viewController = viewController
        self.purchaseInfo = purchaseInfo
        self.spinner = spinner
    }
    
    // MARK: - Progress Dialog Manipulation
    func showProgressDialog(on viewController: UIViewController) {
        progressDialog = try? transaction.getProgressView()
        if let spinner = progressDialog {
            viewController.view.addSubview(spinner)
            viewController.view.bringSubviewToFront(spinner)
            spinner.autoSetDimensions(to: CGSize(width: viewController.view.bounds.size.width, height: viewController.view.bounds.size.height))
            spinner.autoCenterInSuperviewMargins()
            spinner.start()
        }
        
        if let selectedDS = selectedDS,
           let dsLogo = USDKUtil.logo(forDSID: selectedDS.directoryServerID) {
            progressDialog?.setProgressViewImage(dsLogo)
        }
    }
    
    private func hideProgressDialog() {
        progressDialog?.removeFromSuperview()
        progressDialog = nil
    }
    
    // MARK: - AReq/ARes
    func startTransaction() {
        let request = SupportedVersionsRequest(cardID: accoutNumber)
        WebService.shared.retrieveSupportedVersions(with: request, completion: supportedVersionsCompletion)
    }
    
    private func supportedVersionsCompletion(supportedVersionsResponse: SupportedVersionsResponse?, error: Error?) -> Void {
        // WARNING: For now, just use the first DS in the response
        if let err = error {
            viewController.showAlert(for: err)
            progressDialog?.stop()
            spinner?.stopAnimating()
            return
        }
        
        guard let supportedVersionsResponse = supportedVersionsResponse,
              let selectedDS = supportedVersionsResponse.directoryServers.first else {
            viewController.showAlert(message: "Supported Versions Response Directory Servers were empty")
            progressDialog?.stop()
            spinner?.stopAnimating()
            return
        }
        
        do {
            self.selectedDS = selectedDS
            let createSpec = UCreateTransactionSpec(directoryServerID: selectedDS.directoryServerID, messageVersion: selectedDS.messageVersion, cardID: accoutNumber)
            transaction = try UThreeDS2ServiceImpl.shared().createTransaction(with: createSpec)
            
            var params: UAuthenticationRequestParameters!
            if UThreeDS2ServiceImpl.shared().isPromptRequested() {
                transaction.u_doPrompt(viewController) { (wasCancelled) in
                    if wasCancelled {
                        self.viewController.dismiss(animated: true, completion: nil)
                        return
                    }
                    
                    params = try? self.transaction.u_getAuthenticationRequestParameters()
                    self.finishStartTransaction(params: params, selectedDS: selectedDS, supportedVersionsResponse: supportedVersionsResponse)
                }
            }
            else if UThreeDS2ServiceImpl.shared().isBiometricAuthRequested() {
                transaction.u_doBiometricAuthentication { (shouldContinueTransaction, laError, errorMessage) in
                    if shouldContinueTransaction {
                        params = try? self.transaction.u_getAuthenticationRequestParameters()
                        self.finishStartTransaction(params: params, selectedDS: selectedDS, supportedVersionsResponse: supportedVersionsResponse)
                    }
                    else if let err = laError as? LAError,
                            err.code == .userCancel {
                        DispatchQueue.main.async {
                            self.viewController.dismiss(animated: true, completion: nil)
                        }
                    }
                    else {
                        DispatchQueue.main.async {
                            self.viewController.showAlert(withTitle: "Biometric Authentication Failure", message: errorMessage ?? "Somethine went wrong with biometric authentication", showOkButton: true)
                        }
                        
                        return
                    }
                }
            }
            else {
                let params = try transaction.u_getAuthenticationRequestParameters()
                self.finishStartTransaction(params: params, selectedDS: selectedDS, supportedVersionsResponse: supportedVersionsResponse)
            }
        }
        catch {
            viewController.showAlert(for: error)
            progressDialog?.stop()
            spinner?.stopAnimating()
        }
    }
        
    private func finishStartTransaction(params: UAuthenticationRequestParameters, selectedDS: SADirectoryServer, supportedVersionsResponse: SupportedVersionsResponse) {
        let areqParams = AuthenticationRequestParameters(messageVersion: selectedDS.messageVersion, messageCategory: "01", dsIdentifier: selectedDS.directoryServerID, deviceData: params.getDeviceData(), sdkTransactionID: params.getSDKTransactionID(), sdkAppID: params.getSDKAppID(), sdkReferenceNumber: params.getSDKReferenceNumber(), sdkType: "01", supportedVersionsThreeDSServerTransID: selectedDS.threeDSServerTransID, extraAuthParams: supportedVersionsResponse.extraAuthParams, sdkEphemPubKey: params.getSDKEphemeralPublicKey())
        let areq = AReq(purchaseInfo: purchaseInfo, authenticationRequestParameters: areqParams)
        WebService.shared.authenticate(with: areq, completion: authCompletion)
    }
    
    private func authCompletion(ares: AuthenticationResponse?, error: Error?) -> Void {
        if let err = error {
            viewController.showAlert(for: err)
            progressDialog?.stop()
            spinner?.stopAnimating()
            return
        }
        
        guard let ares = ares else {
            viewController.showAlert(message: "ARes was nil")
            progressDialog?.stop()
            spinner?.stopAnimating()
            return
        }
        
        switch ares.transStatus {
        case .challenge:
            let sdkHandler = StoryboardScene.Main.sdkHandlerViewController.instantiate()
            sdkHandler.inject(transaction: transaction, ares: ares, pureInterfaceSCATask: self)
            let navController = UINavigationController(rootViewController: sdkHandler)
            navController.modalPresentationStyle = .fullScreen
            viewController.present(navController, animated: true, completion: nil)
        case .accept, .informationOnly:
            viewController.showAlert(message: "Frictionless auth completed\n\n3DS Server TransID \(ares.threeDSServerTransID ?? "N/A")\n\nStatus: \(ares.transStatus?.rawValue ?? "N/A")")
            cleanUp(with: ares)
        case .decoupledAuthentication:
            viewController.showAlert(message: "Decoupled auth processing\n\n3DS Server TransID \(ares.threeDSServerTransID ?? "N/A")\n\nStatus: \(ares.transStatus?.rawValue ?? "N/A")")
        default:
            viewController.showAlert(message: "Not Authenticated \n\n3DS Server TransID \(ares.threeDSServerTransID ?? "N/A")\n\nStatus: \(ares.transStatus?.rawValue ?? "N/A")")
        }
        
        progressDialog?.stop()
        spinner?.stopAnimating()
    }
    
    // MARK: - Challenges
    func doChallenge(with viewController: UIViewController, challengeStatusReceiver: UChallengeStatusReceiver, ares: AuthenticationResponse, transaction: UTransaction, dismissAction: UIAlertAction) {
        guard let navController = viewController.navigationController else {
            viewController.showAlert(message: "You must present challenge flows within a UINavigationController per EMVco 3DS specification (see what they call a 'tool bar').", actions: [dismissAction])
            return
        }
        
        navController.navigationBar.isTranslucent = false
        guard let threeDSServerTransID = ares.threeDSServerTransID,
                let acsSignedContent = ares.acsSignedContent,
                let acsTransID = ares.acsTransID else {
            viewController.showAlert(message: "3DS Server Trans ID, ACS Signed Content, or ACS Trans ID is missing. Cannot continue the challenge.", actions: [dismissAction])
            return
        }
        
        do {
            let challengeParams = try UChallengeParameters(threeDSTransID: threeDSServerTransID, acsTransactionID: acsTransID, acsRefNumber: ares.acsReferenceNumber, acsSignedContent: acsSignedContent)
            let timeout = Int32(60)
            try transaction.u_doChallenge(navController, challengeParameters: challengeParams, challengeStatusReceiver: challengeStatusReceiver, timeOut: timeout)
        }
        catch {
            print("Failed to do challenge: \(error)")
            viewController.showAlert(message: error.sdkMessage, actions: [dismissAction])
        }
    }
    
    // MARK: - Close
    func cleanUp(with ares: AuthenticationResponse?) {
        hideProgressDialog()
        guard let messageExtensions = ares?.messageExtension else {
            try? transaction.close()
            return
        }

        var exts: [UMessageExtension] = []
        for ext in messageExtensions {
            if let acsRef = ext.data.acsReferenceNumber,
                let x = ext.data.acsEphemPubKey?.x,
                let y = ext.data.acsEphemPubKey?.y,
                let kty = ext.data.acsEphemPubKey?.kty,
                let crv = ext.data.acsEphemPubKey?.crv {
                
                let uExt = UMessageExtension(id: ext.id, name: ext.name, criticalityIndicator: ext.criticalityIndicator, deviceCookie: ext.data.deviceCookie ?? "FIG COOKIE", acsReference: acsRef, bankID: ext.data.bankId ?? "", methods: ext.data.methodsStrings, x: x, y: y, kty: kty, crv: crv, heading: ext.data.usdkPrompt?.heading ?? "Default FIG Heading", message: ext.data.usdkPrompt?.message ?? "Default FIG Message", keyboardType: ext.data.usdkPrompt?.keyboard.uiKeyboardType ?? UIKeyboardType.default)
                exts.append(uExt)
            }
        }
    
        let closeSpec = UCloseTransactionSpec(messageExtensions: exts)
        try? transaction.close(with: closeSpec)
    }
}
