//
//  HomeViewController.swift
//  ThreeDS-Sample
//  Created by Drew Pitchford on 5/14/20.
//  Copyright © 2020 mSignia. All rights reserved.
//

import UIKit
import CoreLocation
import uSDK
import Sentry

class HomeTableViewController: UITableViewController, AppInfoDetailable {
    
    // MARK: - IBOutlets
    @IBOutlet weak var createTransactionButton: UIButton!
    @IBOutlet weak var acctNumberTextField: UITextField!
    @IBOutlet weak var crashReportingTextField: SATextField!
    @IBOutlet weak var paysimEmailTextField: UITextField!
    @IBOutlet weak var paysimUrlTextField: UITextField!
    @IBOutlet weak var versionLabel: UILabel!
    @IBOutlet weak var paysimPasswordTextField: SATextField!
    @IBOutlet weak var usdkLicenseKeyTextField: SATextField!
    @IBOutlet weak var simplifiedInterfaceSwitch: UISwitch!
    @IBOutlet weak var pureInterfaceSwitch: UISwitch!
    @IBOutlet weak var pureInterfaceSCASwitch: UISwitch!
    @IBOutlet var allSwitches: [UISwitch]!
    
    // MARK: - Properties
    private var configParams: UConfigParameters {
        let params = UConfigParameters()
        params.add(USDKUtil.uDirectoryServers())
        return params
    }
    private var purchaseInfo: PurchaseInformation {
        return PurchaseInformation(email: paysimEmailTextField.text ?? "", cardID: acctNumberTextField.text ?? Defaults.acctNumber, purchaseAmount: "100", purchaseCurrency: "840", purchaseExponent: "2", purchaseDate: "20200608094835")
    }
    private var spinner: UIActivityIndicatorView?
    private var simplifiedInterfaceTask: SimplifiedInterfaceTask?
    private var pureInterfaceTask: PureInterfaceTask?
    private var pureInterfaceSCATask: PureInterfaceSCATask?
    
    // MARK: - Lifecycle
    override func viewDidLoad() {
        super.viewDidLoad()
        
        LocationService.shared.requestPermission()
        setUpUI()
        setUpAccessibility()
        
        if let licenseKey = BuildEnvironment.licenseKey,
           !licenseKey.isEmpty  {
            USDKUtil.initSDK(viewController: self, configParams: configParams)
        }
    }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        
        spinner?.stopAnimating()
    }
    
    // MARK: - Setup
    private func setUpUI() {
        // Text Fields
        paysimEmailTextField.text = BuildEnvironment.paysimEmail
        paysimPasswordTextField.text = BuildEnvironment.paysimPassword
        paysimUrlTextField.text = BuildEnvironment.baseURL
        usdkLicenseKeyTextField.text = BuildEnvironment.licenseKey
        acctNumberTextField.text = BuildEnvironment.accountNumber
        crashReportingTextField.text = BuildEnvironment.crashReportingDSN
        
        // Switches
        switch BuildEnvironment.activeAPIInterface {
        case .simplified:
            simplifiedInterfaceSwitch.isOn = true
        case .pure:
            pureInterfaceSwitch.isOn = true
        case .pureSCA:
            pureInterfaceSCASwitch.isOn = true
        }
        
        // Activity Indicator
        if #available(iOS 13.0, *) {
            spinner = UIActivityIndicatorView(style: .large)
        }
        else {
            spinner = UIActivityIndicatorView(style: .whiteLarge)
        }
        
        spinner?.center = CGPoint(x: UIScreen.main.bounds.size.width / 2, y: UIScreen.main.bounds.size.height / 2)
        if let spinner = spinner {
            view.addSubview(spinner)
        }
        
        // Version Label
        versionLabel.text = "app: \(fetchAppVersion() ?? "N/A"), usdk: \(fetchUSDKVersion() ?? "N/A")"
    }
    
    private func setUpAccessibility() {
        acctNumberTextField.accessibilityLabel = "acctNumber"
    }
    
    // MARK: - Helpers
    private func areTextFieldsValid() -> (emailIsValid: Bool, acctNumberIsValid: Bool) {
        return (paysimEmailTextField.isValid(), acctNumberTextField.isValid())
    }
    
    private func updateSwitchStates(with sender: UISwitch) {
        for aSwitch in allSwitches {
            if aSwitch != sender {
                aSwitch.isOn = false
            }
            
            if aSwitch === sender {
                aSwitch.isOn = true
            }
        }
    }
    
    // MARK: - IBActions
    @IBAction func switchWasFlipped(_ sender: UISwitch) {
        updateSwitchStates(with: sender)
        
        switch sender {
        case simplifiedInterfaceSwitch:
            BuildEnvironment.activeAPIInterface = .simplified
        case pureInterfaceSwitch:
            BuildEnvironment.activeAPIInterface = .pure
        case pureInterfaceSCASwitch:
            BuildEnvironment.activeAPIInterface = .pureSCA
        default:
            return
        }
    }
    
    @IBAction func createTransactionWasPressed(_ sender: UIButton) {
        let textFieldValidity = areTextFieldsValid()
        if textFieldValidity.emailIsValid &&
            textFieldValidity.acctNumberIsValid {
            spinner?.startAnimating()
            view.endEditing(true)
            switch BuildEnvironment.activeAPIInterface {
            case .simplified:
                authenticateUser()
            case .pure:
                startTransaction()
            case .pureSCA:
                startTransactionWithSCA()
            }
        }
        else if !textFieldValidity.emailIsValid {
            showAlert(withTitle: "Missing Value", message: "Please enter an email address")
        }
        else if !textFieldValidity.acctNumberIsValid {
            showAlert(withTitle: "Missing Value", message: "Please enter an account number")
        }
    }
    
    @IBAction func simulateCrashWasPressed(_ sender: UIButton) {
        spinner = nil
        spinner!.removeFromSuperview() // this causes a crash
    }
    
    
    // MARK: - Simplified Interface
    private func authenticateUser() {
        if let acctNumber = acctNumberTextField.text {
            simplifiedInterfaceTask = SimplifiedInterfaceTask(viewController: self, delegate: self, purchaseInfo: purchaseInfo, accountNumber: acctNumber)
            simplifiedInterfaceTask?.authenticateUser()
        }
        else {
            showAlert(message: "Account Number was not found. Did you type something into the text field above?")
        }
    }
    
    // MARK: - Pure 3DS Spec Interface
    private func startTransaction() {
        if let acctNumber = acctNumberTextField.text {
            pureInterfaceTask = PureInterfaceTask(accountNumber: acctNumber, viewController: self, purchaseInfo: purchaseInfo, spinner: spinner)
            pureInterfaceTask?.startTransaction()
        }
        else {
            showAlert(message: "Account Number was not found. Did you type something into the text field above?")
        }
    }
    
    private func startTransactionWithSCA() {
        if let acctNumber = acctNumberTextField.text {
            pureInterfaceSCATask = PureInterfaceSCATask(accountNumber: acctNumber, viewController: self, purchaseInfo: purchaseInfo, spinner: spinner)
            pureInterfaceSCATask?.startTransaction()
        }
        else {
            showAlert(message: "Account Number was not found. Did you type something into the text field above?")
        }
    }
}

// MARK: - UITextFieldDelegate methods
extension HomeTableViewController: UITextFieldDelegate {
    func textFieldDidEndEditing(_ textField: UITextField) {
        if let text = textField.text {
            switch textField {
            case paysimEmailTextField:
                BuildEnvironment.paysimEmail = text
                
            case paysimPasswordTextField:
                BuildEnvironment.paysimPassword = text
                
            case paysimUrlTextField:
                BuildEnvironment.baseURL = text
                
            case usdkLicenseKeyTextField:
                BuildEnvironment.licenseKey = text
                USDKUtil.initSDK(viewController: self, configParams: configParams)
                
            case acctNumberTextField:
                BuildEnvironment.accountNumber = text
                
            case crashReportingTextField:
                BuildEnvironment.crashReportingDSN = text
                
            default:
                return
            }
        }
    }
    
    func textFieldShouldReturn(_ textField: UITextField) -> Bool {
        switch textField {
        case crashReportingTextField:
            textField.resignFirstResponder()
        default:
            textField.nextField?.becomeFirstResponder()
        }
        
        return true
    }
}

// MARK: - UAuthenticationDelegate methods
extension HomeTableViewController: UAuthenticationDelegate {
    func authenticated(_ threeDSServerTransID: String, status: UTransStatus) {
        DispatchQueue.main.asyncAfter(deadline: .now() + Defaults.uiDelay) {
            self.spinner?.stopAnimating()
            self.showAlert(message: "The authenticated callback called\n\n3DS Server TransID \(threeDSServerTransID)\n\nStatus: \(status.humanReadableValue)")
            print("User authenticated! \(threeDSServerTransID) \(status)")
        }
    }
    
    func notAuthenticated(_ threeDSServerTransID: String, status: UTransStatus) {
        DispatchQueue.main.asyncAfter(deadline: .now() + Defaults.uiDelay) {
            self.spinner?.stopAnimating()
            self.showAlert(message: "The notAuthenticated callback called\n\n3DS Server TransID \(threeDSServerTransID)\n\nStatus: \(status.humanReadableValue)")
            print("User not authenticated! \(threeDSServerTransID) \(status)")
        }
    }
    
    func decoupledAuthBeingPerformed(_ threeDSServerTransID: String, status: UTransStatus) {
        DispatchQueue.main.asyncAfter(deadline: .now() + Defaults.uiDelay) {
            self.spinner?.stopAnimating()
            self.showAlert(message: "The decoupledAuthBeingPerformed callback called\n\n3DS Server TransID \(threeDSServerTransID)\n\nStatus: \(status.humanReadableValue)")
            print("User decoupled-authed! \(threeDSServerTransID) \(status)")
        }
    }
    
    func cancelled(_ threeDSServerTransID: String) {
        DispatchQueue.main.asyncAfter(deadline: .now() + Defaults.uiDelay) {
            self.spinner?.stopAnimating()
            self.showAlert(message: "The cancelled callback called\n\n3DS Server TransID \(threeDSServerTransID)")
            print("Authentication cancelled! \(threeDSServerTransID)")
        }
    }
    
    func error(_ threeDSServerTransID: String, error: Error) {
        DispatchQueue.main.asyncAfter(deadline: .now() + Defaults.uiDelay) {
            self.spinner?.stopAnimating()
            self.showAlert(message: "The error callback called\n\n3DS Server TransID: \(threeDSServerTransID)\n\n\(error.localizedDescription)")
            print("Error occurred: \(error)")
        }
    }
}
