//
//  SDKHandlerViewController.swift
//  ThreeDS-Sample
//
//  Created by Drew Pitchford on 3/29/21.
//  Copyright © 2021 mSignia. All rights reserved.
//

import UIKit
import uSDK

class SDKHandlerViewController: UIViewController {
    
    // MARK: - Properties
    private var transaction: UTransaction!
    private var ares: AuthenticationResponse!
    private var pureInterfaceSCATask: PureInterfaceSCATask?
    private var pureInterfaceTask: PureInterfaceTask?
    private var dismissAction: UIAlertAction {
        return UIAlertAction(title: "Ok", style: .default) { (_) in
            self.dismiss(animated: true, completion: nil)
        }
    }
    
    // MARK: - Lifecycle
    override func viewDidLoad() {
        super.viewDidLoad()
        
        assertDependenciesHaveBeenInjected([transaction, ares])
    }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        
        if let pureInterfaceTask = pureInterfaceTask {
            pureInterfaceTask.showProgressDialog(on: self)
            pureInterfaceTask.doChallenge(with: self, challengeStatusReceiver: self, ares: ares, transaction: transaction, dismissAction: dismissAction)
        }
        else if let pureInterfaceSCATask = pureInterfaceSCATask {
            pureInterfaceSCATask.showProgressDialog(on: self)
            pureInterfaceSCATask.doChallenge(with: self, challengeStatusReceiver: self, ares: ares, transaction: transaction, dismissAction: dismissAction)
        }
        else {
            showAlert(message: "Oh no! An invalid API Interface task is being used!")
        }
    }
    
    func inject(transaction: UTransaction, ares: AuthenticationResponse, pureInterfaceTask: PureInterfaceTask? = nil, pureInterfaceSCATask: PureInterfaceSCATask? = nil) {
        self.transaction = transaction
        self.ares = ares
        self.pureInterfaceTask = pureInterfaceTask
        self.pureInterfaceSCATask = pureInterfaceSCATask
    }
    
    // MARK: - Helpers
    private func dismissalAlert(withTitle title: String? = nil, message: String? = nil) -> UIAlertController {
        let alert = UIAlertController(title: title, message: message, preferredStyle: .alert)
        alert.addAction(dismissAction)
        return alert
    }
}

// MARK: - UChallengeStatusReceiver methods
extension SDKHandlerViewController: UChallengeStatusReceiver {
    func completed(_ completionEvent: UCompletionEvent) {
        let alert = dismissalAlert(withTitle: "Completed", message: "Challenge was completed\n\nTrans Status:\(completionEvent.getTransactionStatus())\n\n3DS Server Trans ID: \(ares.threeDSServerTransID ?? "N/A")")
        present(alert, animated: true, completion: nil)
        pureInterfaceTask?.cleanUp()
        pureInterfaceSCATask?.cleanUp(with: ares)
    }
    
    func cancelled() {
        let alert = dismissalAlert(withTitle: "Cancelled", message: "Challenge was cancelled by the user")
        present(alert, animated: true, completion: nil)
        pureInterfaceTask?.cleanUp()
        pureInterfaceSCATask?.cleanUp(with: ares)
    }
    
    func timedout() {
        let alert = dismissalAlert(withTitle: "Timedout", message: "Challenge timedout")
        present(alert, animated: true, completion: nil)
        pureInterfaceTask?.cleanUp()
        pureInterfaceSCATask?.cleanUp(with: ares)
    }
    
    func protocolError(_ protocolErrorEvent: UProtocolErrorEvent) {
        let alert = dismissalAlert(withTitle: "Protocol Error", message: "Protocol Error occurred\n\nSDK Trans ID: \(protocolErrorEvent.getSDKTransactionID())\n\nDescription: \(protocolErrorEvent.getErrorMessage().getErrorDescription())\n\nError Code: \(protocolErrorEvent.getErrorMessage().getErrorCode())\n\nDetails: \(protocolErrorEvent.getErrorMessage().getErrorDetails())")
        present(alert, animated: true, completion: nil)
        pureInterfaceTask?.cleanUp()
        pureInterfaceSCATask?.cleanUp(with: ares)
    }
    
    func runtimeError(_ runtimeErrorEvent: URuntimeErrorEvent) {
        let alert = dismissalAlert(withTitle: "Runtime Error", message: "Runtime Error occurred\n\nError Code: \(runtimeErrorEvent.getErrorCode())\n\nMessage: \(runtimeErrorEvent.getErrorMessage())")
        present(alert, animated: true, completion: nil)
        pureInterfaceTask?.cleanUp()
        pureInterfaceSCATask?.cleanUp(with: ares)
    }
}
